# coding: utf-8
# pylint: disable=invalid-name
"""
util functions for 2D EIT
1. generate stimulation lines/patterns
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np


def eit_scan_lines(n_el: int = 16, dist: int = 1) -> np.ndarray:
    """
    Generate scan matrix, `ex_mat` ( or excitation pattern), see notes

    Parameters
    ----------
    n_el : int, optional
        number of electrodes, by default 16
    dist : int, optional
        distance (number of electrodes) of A to B, by default 1
        For 'adjacent'- or 'neighbore'-mode (default) use `1` , and
        for 'apposition'-mode use `n_el/2` (see Examples).

    Returns
    -------
    np.ndarray
        stimulation matrix; shape (n_exc, 2)

    Notes
    -----
        - in the scan of EIT (or stimulation matrix), we use 4-electrodes
        mode, where A, B are used as positive and negative stimulation
        electrodes and M, N are used as voltage measurements.
        - `1` (A) for positive current injection, `-1` (B) for negative current
        sink

    Examples
    --------
        n_el=16
        if mode=='neighbore':
            ex_mat = eit_scan_lines(n_el=n_el)
        elif mode=='apposition':
            ex_mat = eit_scan_lines(dist=n_el/2)

    WARNING
    -------
        `ex_mat` is a local index, where it is ranged from 0...15, within the
        range of the number of electrodes. In FEM applications, you should
        convert `ex_mat` to global index using the (global) `el_pos` parameters.
    """
    return np.array([[i, np.mod(i + dist, n_el)] for i in range(n_el)])


if __name__ == "__main__":
    m = eit_scan_lines(dist=8)
    print(m)



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-locals
# pylint: disable=too-many-instance-attributes
""" 2D/3D FEM routines """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

from typing import Tuple, Union
import numpy as np
import numpy.linalg as la
from scipy import sparse
import warnings
import scipy.sparse.linalg
from pyeit.eit.protocol import PyEITProtocol
from pyeit.mesh import PyEITMesh


class Forward:
    """FEM forward computing code"""

    def __init__(self, mesh: PyEITMesh) -> None:
        """
        FEM forward solver.
        A good FEM forward solver should only depend on
        mesh structure and the position of electrodes.

        Parameters
        ----------
        mesh: PyEITMesh
            mesh object

        Note
        ----
        The nodes are continuous numbered, the numbering of an element is
        CCW (counter-clock-wise).
        """
        self.mesh = mesh
        # coefficient matrix [initialize]
        self.se = calculate_ke(self.mesh.node, self.mesh.element)
        self.assemble_pde(self.mesh.perm)

    def assemble_pde(self, perm: Union[int, float, np.ndarray]) -> None:
        """
        assemble PDE

        Parameters
        ----------
        perm : Union[int, float, np.ndarray]
            permittivity on elements ; shape (n_tri,).
            if `None`, assemble_pde is aborded

        """
        if perm is None:
            return
        perm = self.mesh.get_valid_perm(perm)
        self.kg = assemble(
            self.se, self.mesh.element, perm, self.mesh.n_nodes, ref=self.mesh.ref_el
        )

    def solve(self, ex_line: np.ndarray = None) -> np.ndarray:
        """
        Calculate and compute the potential distribution (complex-valued)
        corresponding to the permittivity distribution `perm ` for a
        excitation contained specified by `ex_line` (Neumann BC)

        Parameters
        ----------
        ex_line : np.ndarray, optional
            stimulation/excitation matrix, of shape (2,)

        Returns
        -------
        np.ndarray
            potential on nodes ; shape (n_pts,)

        Notes
        -----
        Currently, only simple electrode model is supported,
        CEM (complete electrode model) is under development.
        """
        # using natural boundary conditions
        b = np.zeros(self.mesh.n_nodes)
        b[self.mesh.el_pos[ex_line]] = [1, -1]

        # solve
        return scipy.sparse.linalg.spsolve(self.kg, b)


class EITForward(Forward):
    """EIT Forward simulation, depends on mesh and protocol"""

    def __init__(self, mesh: PyEITMesh, protocol: PyEITProtocol) -> None:
        """
        EIT Forward Solver

        Parameters
        ----------
        mesh: PyEITMesh
            mesh object
        protocol: PyEITProtocol
            measurement object

        Notes
        -----
        The Jacobian and the boundary voltages used the SIGN information,
        for example, V56 = V6 - V5 = -V65. If you are using absolute boundary
        voltages for imaging, you MUST normalize it with the signs of v0
        under each current-injecting pattern.
        """
        self._check_mesh_protocol_compatibility(mesh, protocol)

        # FEM solver
        super().__init__(mesh=mesh)

        # EIT measurement protocol
        self.protocol = protocol

    def _check_mesh_protocol_compatibility(
        self, mesh: PyEITMesh, protocol: PyEITProtocol
    ) -> None:
        """
        Check if mesh and protocol are compatible

        - #1 n_el in mesh >=  n_el in protocol
        - #2 .., TODO if necessary

        Raises
        ------
        ValueError
            if protocol is not compatible to the mesh
        """
        # n_el in mesh should be >=  n_el in protocol
        m_n_el = mesh.n_el
        p_n_el = protocol.n_el

        if m_n_el != p_n_el:
            warnings.warn(
                f"The mesh use {m_n_el} electrodes, and the protocol use only {p_n_el} electrodes",
                stacklevel=2,
            )

        if m_n_el < p_n_el:
            raise ValueError(
                f"Protocol is not compatible with mesh :\
The mesh use {m_n_el} electrodes, and the protocol use only {p_n_el} electrodes "
            )

    def solve_eit(
        self,
        perm: Union[int, float, np.ndarray] = None,
    ) -> np.ndarray:
        """
        EIT simulation, generate forward v measurements

        Parameters
        ----------
        perm : Union[int, float, np.ndarray], optional
            permittivity on elements ; shape (n_tri,), by default `None`.
            if perm is `None`, the computation of forward v measurements will be
            based on the permittivity of the mesh, self.mesh.perm
        Returns
        -------
        v: np.ndarray
            simulated boundary voltage measurements; shape(n_exe*n_el,)
        """
        self.assemble_pde(perm)
        v = np.zeros((self.protocol.n_exc, self.protocol.n_meas))
        for i, ex_line in enumerate(self.protocol.ex_mat):
            f = self.solve(ex_line)
            v[i] = subtract_row(f[self.mesh.el_pos], self.protocol.meas_mat[i])

        return v.reshape(-1)

    def compute_jac(
        self,
        perm: Union[int, float, np.ndarray] = None,
        normalize: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the Jacobian matrix and initial boundary voltage meas.
        extimation v0

        Parameters
        ----------
        perm : Union[int, float, np.ndarray], optional
            permittivity on elements ; shape (n_tri,), by default `None`.
            if perm is `None`, the computation of Jacobian matrix will be based
            on the permittivity of the mesh, self.mesh.perm
        normalize : bool, optional
            flag for Jacobian normalization, by default False.
            If True the Jacobian is normalized

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Jacobian matrix, initial boundary voltage meas. extimation v0

        """
        # update k if necessary and calculate r=inv(k)
        self.assemble_pde(perm)
        r_el = la.inv(self.kg.toarray())[self.mesh.el_pos]

        # calculate v, jac per excitation pattern (ex_line)
        _jac = np.zeros(
            (self.protocol.n_exc, self.protocol.n_meas, self.mesh.n_elems),
            dtype=self.mesh.perm.dtype,
        )
        v = np.zeros((self.protocol.n_exc, self.protocol.n_meas))
        for i, ex_line in enumerate(self.protocol.ex_mat):
            f = self.solve(ex_line)
            v[i] = subtract_row(f[self.mesh.el_pos], self.protocol.meas_mat[i])
            ri = subtract_row(r_el, self.protocol.meas_mat[i])
            # Build Jacobian matrix column wise (element wise)
            #    Je = Re*Ke*Ve = (nex3) * (3x3) * (3x1)
            for (e, ijk) in enumerate(self.mesh.element):
                _jac[i, :, e] = np.dot(np.dot(ri[:, ijk], self.se[e]), f[ijk])

        # measurement protocol
        jac = np.vstack(_jac)
        v0 = v.reshape(-1)

        # Jacobian normalization: divide each row of J (J[i]) by abs(v0[i])
        if normalize:
            jac = jac / np.abs(v0[:, None])
        return jac, v0

    def compute_b_matrix(
        self,
        perm: Union[int, float, np.ndarray] = None,
    ) -> np.ndarray:
        """
        Compute back-projection mappings (smear matrix)

        Parameters
        ----------
        perm : Union[int, float, np.ndarray], optional
            permittivity on elements ; shape (n_tri,), by default `None`.
            if perm is `None`, the computation of smear matrix will be based
            on the permittivity of the mesh, self.mesh.perm

        Returns
        -------
        np.ndarray
            back-projection mappings (smear matrix); shape(n_exc, n_pts, 1), dtype= bool
        """
        self.assemble_pde(perm)
        b_mat = np.zeros((self.protocol.n_exc, self.protocol.n_meas, self.mesh.n_nodes))

        for i, ex_line in enumerate(self.protocol.ex_mat):
            f = self.solve(ex_line=ex_line)
            f_el = f[self.mesh.el_pos]
            # build bp projection matrix
            # 1. we can either smear at the center of elements, using
            #    >> fe = np.mean(f[:, self.tri], axis=1)
            # 2. or, simply smear at the nodes using f
            b_mat[i] = _smear(f, f_el, self.protocol.meas_mat[i])

        return np.vstack(b_mat)


def _smear(f: np.ndarray, fb: np.ndarray, pairs: np.ndarray) -> np.ndarray:
    """
    Build smear matrix B for bp for one exitation

    used for the smear matrix computation by @ChabaneAmaury

    Parameters
    ----------
    f: np.ndarray
        potential on nodes
    fb: np.ndarray
        potential on adjacent electrodes
    pairs: np.ndarray
        electrodes numbering pairs

    Returns
    -------
    B: np.ndarray
        back-projection matrix
    """
    # Replacing the code below by a faster implementation in Numpy
    f_min = np.minimum(fb[pairs[:, 0]], fb[pairs[:, 1]]).reshape((-1, 1))
    f_max = np.maximum(fb[pairs[:, 0]], fb[pairs[:, 1]]).reshape((-1, 1))
    return (f_min < f) & (f <= f_max)


def smear_nd(
    f: np.ndarray, fb: np.ndarray, meas_pattern: np.ndarray, new: bool = False
) -> np.ndarray:
    """
    Build smear matrix B for bp

    Parameters
    ----------
    f: np.ndarray
        potential on nodes; shape (n_exc, n_pts)
    fb: np.ndarray
        potential on adjacent electrodes; shape (n_exc, n_el)
    meas_pattern: np.ndarray
        electrodes numbering pairs; shape (n_exc, n_meas, 2)
    new : bool, optional
        flag to use new matrices based computation, by default False.
        If `False` to smear-computation from ChabaneAmaury is used

    Returns
    -------
    np.ndarray
        back-projection (smear) matrix; shape (n_exc, n_meas, n_pts), dtype= bool
    """
    if new:
        # new implementation not much faster! :(
        idx_meas_0 = meas_pattern[:, :, 0]
        idx_meas_1 = meas_pattern[:, :, 1]
        n_exc = meas_pattern.shape[0]  # number of excitations
        n_meas = meas_pattern.shape[1]  # number of measurements per excitations
        n_pts = f.shape[1]  # number of nodes
        idx_exc = np.ones_like(idx_meas_0, dtype=int) * np.arange(n_exc).reshape(
            n_exc, 1
        )
        f_min = np.minimum(fb[idx_exc, idx_meas_0], fb[idx_exc, idx_meas_1])
        f_max = np.maximum(fb[idx_exc, idx_meas_0], fb[idx_exc, idx_meas_1])
        # contruct matrices of shapes (n_exc, n_meas, n_pts) for comparison
        f_min = np.repeat(f_min[:, :, np.newaxis], n_pts, axis=2)
        f_max = np.repeat(f_max[:, :, np.newaxis], n_pts, axis=2)
        f0 = np.repeat(f[:, :, np.newaxis], n_meas, axis=2)
        f0 = f0.swapaxes(1, 2)
        return (f_min < f0) & (f0 <= f_max)
    else:
        # Replacing the below code by a faster implementation in Numpy
        def b_matrix_init(k):
            return _smear(f[k], fb[k], meas_pattern[k])

        return np.array(list(map(b_matrix_init, np.arange(f.shape[0]))))


def subtract_row(v: np.ndarray, meas_pattern: np.ndarray) -> np.ndarray:
    """
    Build the voltage differences on axis=1 using the meas_pattern.
    v_diff[k] = v[i, :] - v[j, :]

    New implementation 33% less computation time

    Parameters
    ----------
    v: np.ndarray
        Nx1 boundary measurements vector or NxM matrix; shape (n_exc,n_el,1)
    meas_pattern: np.ndarray
        Nx2 subtract_row pairs; shape (n_exc, n_meas, 2)

    Returns
    -------
    np.ndarray
        difference measurements v_diff
    """
    return v[meas_pattern[:, 0]] - v[meas_pattern[:, 1]]


def assemble(
    ke: np.ndarray, tri: np.ndarray, perm: np.ndarray, n_pts: int, ref: int = 0
) -> np.ndarray:
    """
    Assemble the stiffness matrix (using sparse matrix)

    Parameters
    ----------
    ke: np.ndarray
        n_tri x (n_dim x n_dim) 3d matrix
    tri: np.ndarray
        the structure of mesh
    perm: np.ndarray
        n_tri x 1 conductivities on elements
    n_pts: int
        number of nodes
    ref: int, optional
        reference electrode, by default 0

    Returns
    -------
    np.ndarray
        NxN array of complex stiffness matrix

    Notes
    -----
    you may use sparse matrix (IJV) format to automatically add the local
    stiffness matrix to the global matrix.
    """
    n_tri, n_vertices = tri.shape

    # New: use IJV indexed sparse matrix to assemble K (fast, prefer)
    # index = np.array([np.meshgrid(no, no, indexing='ij') for no in tri])
    # note: meshgrid is slow, using handcraft sparse index, for example
    # let tri=[[1, 2, 3], [4, 5, 6]], then indexing='ij' is equivalent to
    # row = [1, 1, 1, 2, 2, 2, ...]
    # col = [1, 2, 3, 1, 2, 3, ...]
    row = np.repeat(tri, n_vertices).ravel()
    col = np.repeat(tri, n_vertices, axis=0).ravel()
    data = np.array([ke[i] * perm[i] for i in range(n_tri)]).ravel()

    # set reference nodes before constructing sparse matrix
    if 0 <= ref < n_pts:
        dirichlet_ind = np.logical_or(row == ref, col == ref)
        # K[ref, :] = 0, K[:, ref] = 0
        row = row[~dirichlet_ind]
        col = col[~dirichlet_ind]
        data = data[~dirichlet_ind]
        # K[ref, ref] = 1.0
        row = np.append(row, ref)
        col = np.append(col, ref)
        data = np.append(data, 1.0)

    # for efficient sparse inverse (csc)
    return sparse.csr_matrix((data, (row, col)), shape=(n_pts, n_pts))


def calculate_ke(pts: np.ndarray, tri: np.ndarray) -> np.ndarray:
    """
    Calculate local stiffness matrix on all elements.

    Parameters
    ----------
    pts: np.ndarray
        Nx2 (x,y) or Nx3 (x,y,z) coordinates of points
    tri: np.ndarray
        Mx3 (triangle) or Mx4 (tetrahedron) connectivity of elements

    Returns
    -------
    np.ndarray
        n_tri x (n_dim x n_dim) 3d matrix
    """
    n_tri, n_vertices = tri.shape

    # check dimension
    # '3' : triangles
    # '4' : tetrahedrons
    if n_vertices == 3:
        _k_local = _k_triangle
    elif n_vertices == 4:
        _k_local = _k_tetrahedron
    else:
        raise TypeError("The num of vertices of elements must be 3 or 4")

    # default data types for ke
    ke_array = np.zeros((n_tri, n_vertices, n_vertices))
    for ei in range(n_tri):
        no = tri[ei, :]
        xy = pts[no]

        # compute the KIJ (permittivity=1.)
        ke = _k_local(xy)
        ke_array[ei] = ke

    return ke_array


def _k_triangle(xy: np.ndarray) -> np.ndarray:
    """
    Given a point-matrix of an element, solving for Kij analytically
    using barycentric coordinates (simplex coordinates)

    Parameters
    ----------
    xy: np.ndarray
        (x,y) of nodes 1,2,3 given in counterclockwise manner

    Returns
    -------
    np.ndarray
        local stiffness matrix
    """
    # edges (vector) of triangles
    s = xy[[2, 0, 1]] - xy[[1, 2, 0]]

    # area of triangles. Note, abs is removed since version 2020,
    # user must make sure all triangles are CCW (conter clock wised).
    # at = 0.5 * np.linalg.det(s[[0, 1]])
    at = 0.5 * det2x2(s[0], s[1])

    # Local stiffness matrix (e for element)
    return np.dot(s, s.T) / (4.0 * at)


def det2x2(s1: np.ndarray, s2: np.ndarray) -> float:
    """Calculate the determinant of a 2x2 matrix"""
    return s1[0] * s2[1] - s1[1] * s2[0]


def _k_tetrahedron(xy: np.ndarray) -> np.ndarray:
    """
    Given a point-matrix of an element, solving for Kij analytically
    using barycentric coordinates (simplex coordinates)

    Parameters
    ----------
    xy: np.ndarray
        (x,y) of nodes 1, 2, 3, 4 given in counterclockwise manner, see notes.

    Returns
    -------
    np.ndarray
        local stiffness matrix

    Notes
    -----
    A tetrahedron is described using [0, 1, 2, 3] (local node index) or
    [171, 27, 9, 53] (global index). Counterclockwise (CCW) is defined
    such that the barycentric coordinate of face (1->2->3) is positive.
    """
    s = xy[[2, 3, 0, 1]] - xy[[1, 2, 3, 0]]

    # volume of the tetrahedron, Note abs is removed since version 2020,
    # user must make sure all tetrahedrons are CCW (counter clock wised).
    vt = 1.0 / 6 * la.det(s[[0, 1, 2]])

    # calculate area (vector) of triangle faces
    # re-normalize using alternative (+,-) signs
    ij_pairs = [[0, 1], [1, 2], [2, 3], [3, 0]]
    signs = [1, -1, 1, -1]
    a = np.array([sign * np.cross(s[i], s[j]) for (i, j), sign in zip(ij_pairs, signs)])

    # local (e for element) stiffness matrix
    return np.dot(a, a.transpose()) / (36.0 * vt)


if __name__ == "__main__":
    """"""



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-arguments
# pylint: disable=too-many-instance-attributes, too-many-locals
# pylint: disable=arguments-differ
""" dynamic EIT solver using JAC """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

from typing import Union

import numpy as np
import scipy.linalg as la

from .base import EitBase


class JAC(EitBase):
    """A sensitivity-based EIT imaging class"""

    def setup(
        self,
        p: float = 0.20,
        lamb: float = 0.001,
        method: str = "kotre",
        jac_normalized: bool = False,
    ) -> None:
        """
        Setup JAC solver

        Jacobian matrix based reconstruction.

        Parameters
        ----------
        p : float, optional
            JAC parameters, by default 0.20
        lamb : float, optional
            JAC parameters, by default 0.001
        method : str, optional
            regularization methods ("kotre", "lm", "dgn" ), by default "kotre"
        jac_normalized : bool, optional
            normalize the jacobian using f0 computed from input perm, by
            default False
        """
        # passing imaging parameters
        self.params = {
            "p": p,
            "lamb": lamb,
            "method": method,
            "jac_normalize": jac_normalized,
        }
        # pre-compute H0 for dynamical imaging
        # H = (J.T*J + R)^(-1) * J.T
        self.J, self.v0 = self.fwd.compute_jac(normalize=jac_normalized)
        self.H = self._compute_h(self.J, p, lamb, method)
        self.is_ready = True

    def _compute_h(
        self, jac: np.ndarray, p: float, lamb: float, method: str = "kotre"
    ) -> np.ndarray:
        """
        Compute self.H matrix for JAC solver

        JAC method of dynamic EIT solver:
            H = (J.T*J + lamb*R)^(-1) * J.T

        Parameters
        ----------
        jac : np.ndarray
            Jacobian
        p : float
            regularization parameter
        lamb : float
            regularization parameter
        method : str, optional
            regularization method, ("kotre", "lm", "dgn" ), by default "kotre"

        Returns
        -------
        np.ndarray
            H matrix, pseudo-inverse matrix of JAC
        """
        j_w_j = np.dot(jac.transpose(), jac)
        if method == "kotre":
            # see adler-dai-lionheart-2007
            # p=0   : noise distribute on the boundary ('dgn')
            # p=0.5 : noise distribute on the middle
            # p=1   : noise distribute on the center ('lm')
            r_mat = np.diag(np.diag(j_w_j)) ** p
        elif method == "lm":
            # Marquardt–Levenberg, 'lm' for short
            # or can be called NOSER, DLS
            r_mat = np.diag(np.diag(j_w_j))
        else:
            # Damped Gauss Newton, 'dgn' for short
            r_mat = np.eye(jac.shape[1])

        # build H
        return np.dot(la.inv(j_w_j + lamb * r_mat), jac.transpose())

    def solve_gs(self, v1: np.ndarray, v0: np.ndarray) -> np.ndarray:
        """
        Solving by weighted frequency

        Parameters
        ----------
        v1: np.ndarray
            current frame
        v0: np.ndarray
            referenced frame

        Raises
        ------
        SolverNotReadyError
            raised if solver not ready (see self._check_solver_is_ready())

        Returns
        -------
        np.ndarray
            complex-valued np.ndarray, changes of conductivities
        """
        self._check_solver_is_ready()
        a = np.dot(v1, v0) / np.dot(v0, v0)
        dv = v1 - a * v0
        # return ds average epsilon on element
        return -np.dot(self.H, dv.transpose())

    def jt_solve(
        self, v1: np.ndarray, v0: np.ndarray, normalize: bool = True
    ) -> np.ndarray:
        """
        a 'naive' back projection using the transpose of Jac.
        This scheme is the one published by kotre (1989):

        Parameters
        ----------
        v1: np.ndarray
            current frame
        v0: np.ndarray
            referenced frame
        normalize : bool, optional
            flag to log-normalize the current frame difference dv, by default
            True. The input (dv) and output (ds) is log-normalized.

        Raises
        ------
        SolverNotReadyError
            raised if solver not ready (see self._check_solver_is_ready())

        Returns
        -------
        np.ndarray
            complex-valued np.ndarray, changes of conductivities

        Notes
        -----
            [1] Kotre, C. J. (1989).
                A sensitivity coefficient method for the reconstruction of
                electrical impedance tomograms.
                Clinical Physics and Physiological Measurement,
                10(3), 275–281. doi:10.1088/0143-0815/10/3/008

        """
        self._check_solver_is_ready()
        if normalize:
            dv = np.log(np.abs(v1) / np.abs(v0)) * np.sign(v0.real)
        else:
            dv = v1 - v0
        # s_r = J^Tv_r
        ds = -np.dot(self.J.conj().T, dv)
        return np.exp(ds) - 1.0

    def gn(
        self,
        v: np.ndarray,
        x0: Union[int, float, np.ndarray] = None,
        maxiter: int = 1,
        gtol: float = 1e-4,
        p: float = None,
        lamb: float = None,
        lamb_decay: float = 1.0,
        lamb_min: float = 0.0,
        method: str = "kotre",
        verbose: bool = False,
        **kwargs,
    ) -> np.ndarray:
        """
        Gaussian Newton Static Solver
        You can use a different p, lamb other than the default ones in setup

        Parameters
        ----------
        v : np.ndarray
            boundary measurement
        x0 : Union[int, float, np.ndarray], optional
            initial permittivity guess, by default None
            (see Foward._get_perm for more details, in fem.py)
        maxiter : int, optional
            number of maximum iterations, by default 1
        gtol : float, optional
            convergence threshold, by default 1e-4
        p : float, optional
            JAC parameters (can be overridden), by default None
        lamb : float, optional
            JAC parameters (can be overridden), by default None
        lamb_decay : float, optional
            decay of lamb0, i.e., lamb0 = lamb0 * lamb_delay of each iteration,
            by default 1.0
        lamb_min : float, optional
            minimal value of lamb, by default 0.0
        method : str, optional
            regularization methods ("kotre", "lm", "dgn" ), by default "kotre"
        verbose : bool, optional
            verbose flag, by default False

        Raises
        ------
        SolverNotReadyError
            raised if solver not ready (see self._check_solver_is_ready())

        Returns
        -------
        np.ndarray
            Complex-valued conductivities, sigma

        Note
        ----
        Gauss-Newton Iterative solver,
            x1 = x0 - (J^TJ + lamb*R)^(-1) * r0
        where:
            R = diag(J^TJ)**p
            r0 (residual) = real_measure - forward_v
        """
        self._check_solver_is_ready()
        if x0 is None:
            x0 = self.mesh.perm
        if p is None:
            p = self.params["p"]
        if lamb is None:
            lamb = self.params["lamb"]
        if method is None:
            method = self.params["method"]

        # convergence test
        x0_norm = np.linalg.norm(x0)

        for i in range(maxiter):

            # forward solver,
            jac, v0 = self.fwd.compute_jac(x0)
            # Residual
            r0 = v - v0

            # Damped Gaussian-Newton
            h_mat = self._compute_h(jac, p, lamb, method)

            # update
            d_k = np.dot(h_mat, r0)
            x0 = x0 - d_k

            # convergence test
            c = np.linalg.norm(d_k) / x0_norm
            if c < gtol:
                break

            if verbose:
                print("iter = %d, lamb = %f, gtol = %f" % (i, lamb, c))

            # update regularization parameter
            # lambda can be given in user defined decreasing lists
            lamb *= lamb_decay
            lamb = max(lamb, lamb_min)
        return x0

    def project(self, ds: np.ndarray) -> np.ndarray:
        """
        Project ds using spatial difference filter (deprecated)

        Parameters
        ----------
        ds : np.ndarray
            delta sigma (conductivities)

        Returns
        -------
        np.ndarray
            _description_
        """
        """project ds using spatial difference filter (deprecated)

        Parameters
        ----------
        ds: np.ndarray
            delta sigma (conductivities)

        Returns
        -------
        np.ndarray
        """
        d_mat = sar(self.mesh.element)
        return np.dot(d_mat, ds)


def h_matrix(
    jac: np.ndarray, p: float, lamb: float, method: str = "kotre"
) -> np.ndarray:
    """
    (NOT USED in JAC solver)
    JAC method of dynamic EIT solver:
        H = (J.T*J + lamb*R)^(-1) * J.T

    Parameters
    ----------
    jac : np.ndarray
        Jacobian
    p : float
        regularization parameter
    lamb : float
        regularization parameter
    method : str, optional
        regularization method, ("kotre", "lm", "dgn" ), by default "kotre"

    Returns
    -------
    np.ndarray
        H matrix, pseudo-inverse matrix of JAC
    """
    j_w_j = np.dot(jac.transpose(), jac)
    if method == "kotre":
        # see adler-dai-lionheart-2007
        # p=0   : noise distribute on the boundary ('dgn')
        # p=0.5 : noise distribute on the middle
        # p=1   : noise distribute on the center ('lm')
        r_mat = np.diag(np.diag(j_w_j)) ** p
    elif method == "lm":
        # Marquardt–Levenberg, 'lm' for short
        # or can be called NOSER, DLS
        r_mat = np.diag(np.diag(j_w_j))
    else:
        # Damped Gauss Newton, 'dgn' for short
        r_mat = np.eye(jac.shape[1])

    # build H
    return np.dot(la.inv(j_w_j + lamb * r_mat), jac.transpose())


def sar(el2no: np.ndarray) -> np.ndarray:
    """
    Extract spatial difference matrix on the neighbors of each element
    in 2D fem using triangular mesh.

    Parameters
    ----------
    el2no : np.ndarray
        triangle structures

    Returns
    -------
    np.ndarray
        SAR matrix
    """
    ne = el2no.shape[0]
    d_mat = np.eye(ne)
    for i in range(ne):
        ei = el2no[i, :]
        #
        i0 = np.argwhere(el2no == ei[0])[:, 0]
        i1 = np.argwhere(el2no == ei[1])[:, 0]
        i2 = np.argwhere(el2no == ei[2])[:, 0]
        idx = np.unique(np.hstack([i0, i1, i2]))
        # build row-i
        for j in idx:
            d_mat[i, j] = -1
        nn = idx.size - 1
        d_mat[i, i] = nn
    return d_mat



# coding: utf-8
# pylint: disable=invalid-name, no-member, arguments-differ
""" bp (back-projection) and f(filtered)-bp module """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np
from .base import EitBase


class BP(EitBase):
    """A naive inversion of (Euclidean) back projection."""

    def setup(self, weight: str = "none") -> None:
        """
        Setup BP solver

        Parameters
        ----------
        weight : str, optional
            BP parameter, by default "none"
        """
        self.params = {"weight": weight}

        # build the weighting matrix
        # BP: in node imaging, H is the smear matrix (transpose of B)
        self.B = self.fwd.compute_b_matrix()
        self.H = self._compute_h(b_matrix=self.B)
        self.is_ready = True

    def _compute_h(self, b_matrix: np.ndarray) -> np.ndarray:
        """
        Compute H matrix for BP solver

        Parameters
        ----------
        b_matrix : np.ndarray
            BP matrix

        Returns
        -------
        np.ndarray
            H matrix
        """
        if self.params["weight"] == "simple":
            weights = self._simple_weight(b_matrix.shape[0])
            b_matrix = weights * b_matrix
        return b_matrix.T

    def solve_gs(self, v1: np.ndarray, v0: np.ndarray) -> np.ndarray:
        """
        Solving using gram-schmidt orthogonalization

        Parameters
        ----------
        v1: np.ndarray
            current frame
        v0: np.ndarray
            referenced frame

        Raises
        ------
        SolverNotReadyError
            raised if solver not ready (see self._check_solver_is_ready())

        Returns
        -------
        np.ndarray
            complex-valued np.ndarray, changes of conductivities
        """
        self._check_solver_is_ready()
        a = np.dot(v1, v0) / np.dot(v0, v0)
        vn = -(v1 - a * v0) / np.sign(v0.real)
        return np.dot(self.H, vn.transpose())

    def _normalize(self, v1: np.ndarray, v0: np.ndarray) -> np.ndarray:
        """
        redefine normalize for BP (without amplitude normalization) using
        only the sign of v0.real. [experimental]

        Normalize current frame using the amplitude of the reference frame.
        Boundary measurements v are complex-valued

        Parameters
        ----------
        v1: np.ndarray
            current frame
        v0: np.ndarray
            referenced frame

        Returns
        -------
        np.ndarray
            Normalized current frame difference dv
        """
        return (v1 - v0) / np.sign(v0.real)

    def _simple_weight(self, num_voltages: int) -> np.ndarray:
        """
        Build weighting matrix : simple, normalize by radius.

        Parameters
        ----------
        num_voltages : int
            number of equal-potential lines

        Returns
        -------
        np.ndarray
            weighting matrix

        Notes
        -----
        as in fem.py, we could either smear at,

        (1) elements, using the center co-ordinates (x,y) of each element
            >> center_e = np.mean(self.pts[self.tri], axis=1)
        (2) nodes.
        """
        d = np.sqrt(np.sum(self.mesh.node**2, axis=1))
        r = np.max(d)
        w = (1.01 * r - d) / (1.01 * r)
        # weighting by element-wise multiplication W with B
        return np.dot(np.ones((num_voltages, 1)), w.reshape(1, -1))



# coding: utf-8
# pylint: disable=invalid-name, no-member, arguments-differ
""" EIT protocol """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import absolute_import, division, print_function

from dataclasses import dataclass
from typing import Union

import numpy as np


@dataclass
class PyEITProtocol:
    """
    EIT Protocol buid-in protocol object

    """

    ex_mat: np.ndarray
    meas_mat: np.ndarray

    def __post_init__(self) -> None:
        """Checking of the inputs"""
        self.ex_mat = self._check_ex_mat(self.ex_mat)
        self.meas_mat = self._check_meas_mat(self.meas_mat)

    def _check_ex_mat(self, ex_mat: np.ndarray) -> np.ndarray:
        """
        Check/init stimulation

        Parameters
        ----------
        ex_mat : np.ndarray
            stimulation/excitation matrix, of shape (n_exc, 2).
            If single stimulation (ex_line) is passed only a list of length 2
            and np.ndarray of size 2 will be treated.

        Returns
        -------
        np.ndarray
            stimulation matrix

        Raises
        ------
        TypeError
            Only accept, list of length 2, np.ndarray of size 2,
            or np.ndarray of shape (n_exc, 2)
        """
        if isinstance(ex_mat, list) and len(ex_mat) == 2:
            # case ex_line has been passed instead of ex_mat
            ex_mat = np.array([ex_mat]).reshape((1, 2))  # build a 2D array
        elif isinstance(ex_mat, np.ndarray) and ex_mat.size == 2:
            #     case ex_line np.ndarray has been passed instead of ex_mat
            ex_mat = ex_mat.reshape((-1, 2))

        if (
            not isinstance(ex_mat, np.ndarray)
            or ex_mat.ndim != 2
            or ex_mat.shape[1] != 2
        ):
            raise TypeError(
                f"Wrong shape of {ex_mat=} expected an ndarray ; shape (n_exc, 2)"
            )

        return ex_mat

    def _check_meas_mat(self, meas_mat: np.ndarray) -> np.ndarray:
        """
        Check measurement pattern

        Parameters
        ----------
        n_exc : int
            number of excitations/stimulations
        meas_pattern : np.ndarray, optional
           measurements pattern / subtract_row pairs [N, M] to check; shape (n_exc, n_meas_per_exc, 2)

        Returns
        -------
        np.ndarray
            measurements pattern / subtract_row pairs [N, M]; shape (n_exc, n_meas_per_exc, 2)

        Raises
        ------
        TypeError
            raised if meas_pattern is not a np.ndarray of shape (n_exc, : , 2)
        """
        if not isinstance(meas_mat, np.ndarray):
            raise TypeError(
                f"Wrong type of {meas_mat=}, expected an ndarray; shape ({self.n_exc}, n_meas_per_exc, 2)"
            )
        # test shape is something like (n_exc, :, 2)
        if meas_mat.ndim != 3 or meas_mat.shape[::2] != (self.n_exc, 2):
            raise TypeError(
                f"Wrong shape of {meas_mat=}: {meas_mat.shape=}, expected an ndarray; shape ({self.n_exc}, n_meas_per_exc, 2)"
            )

        return meas_mat

    @property
    def n_exc(self) -> int:
        """
        Returns
        -------
        int
            number of excitation
        """
        return self.ex_mat.shape[0]

    @property
    def n_meas(self) -> int:
        """
        Returns
        -------
        int
            number of measurements per excitations
        """
        return self.meas_mat.shape[1]

    @property
    def n_meas_tot(self) -> int:
        """
        Returns
        -------
        int
            total amount of measurements
        """
        return self.n_meas * self.n_exc

    @property
    def n_el(self) -> int:
        """
        Returns
        -------
        int
            number of electrodes used in the excitation and
        """
        return max(max(self.ex_mat.flatten()), max(self.meas_mat.flatten())) + 1


def create(
    n_el: int = 16,
    dist_exc: Union[int, list[int]] = 1,
    step_meas: int = 1,
    parser_meas: Union[str, list[str]] = "std",
) -> PyEITProtocol:
    """
    Return an EIT protocol, comprising an excitation and a measuremnet pattern

    Parameters
    ----------
    n_el : int, optional
        number of total electrodes, by default 16
    dist_exc : Union[int, list[int]], optional
        distance (number of electrodes) of A to B, by default 1
        For 'adjacent'- or 'neighbore'-mode (default) use `1` , and
        for 'apposition'-mode use `n_el/2`. (see `build_exc_pattern`)
        if a list of integer is passed the excitation will bee stacked together.
    step_meas : int, optional
    measurement method (two adjacent electrodes are used for measuring), by default 1 (adjacent).
        (see `build_meas_pattern`)
    parser_meas : Union[str, list[str]], optional
        parsing the format of each frame in measurement/file, by default 'std'.
        (see `build_meas_pattern`)

    Returns
    -------
    PyEITProtocol
        EIT protocol object

    Raises
    ------
    TypeError
        if dist_exc is not list or an int
    """
    if isinstance(dist_exc, int):
        dist_exc = [dist_exc]

    if not isinstance(dist_exc, list):
        raise TypeError(f"{dist_exc=}; {type(dist_exc)=} should be a list[int]")

    _ex_mat = [build_exc_pattern_std(n_el, dist) for dist in dist_exc]
    ex_mat = np.vstack(_ex_mat)

    meas_mat = build_meas_pattern_std(ex_mat, n_el, step_meas, parser_meas)
    return PyEITProtocol(ex_mat, meas_mat)


def build_meas_pattern_std(
    ex_mat: np.ndarray,
    n_el: int = 16,
    step: int = 1,
    parser: Union[str, list[str]] = "std",
) -> np.ndarray:
    """
    Build the measurement pattern (subtract_row-voltage pairs [N, M])
    for all excitations on boundary electrodes.

    we direct operate on measurements or Jacobian on electrodes,
    so, we can use LOCAL index in this module, do not require el_pos.

    Notes
    -----
    ABMN Model.
    A: current driving electrode,
    B: current sink,
    M, N: boundary electrodes, where v_diff = v_n - v_m.

    Parameters
    ----------
    ex_mat : np.ndarray
        Nx2 array, [positive electrode, negative electrode]. ; shape (n_exc, 2)
    n_el : int, optional
        number of total electrodes, by default 16
    step : int, optional
        measurement method (two adjacent electrodes are used for measuring), by default 1 (adjacent)
    parser : Union[str, list[str]], optional
        parsing the format of each frame in measurement/file, by default 'std'
        if parser contains 'fmmu', or 'rotate_meas' then data are trimmed,
        boundary voltage measurements are re-indexed and rotated,
        start from the positive stimulus electrode start index 'A'.
        if parser contains 'std', or 'no_rotate_meas' then data are trimmed,
        the start index (i) of boundary voltage measurements is always 0.
        if parser contains 'meas_current', the measurements on current carrying
        electrodes are allowed. Otherwise the measurements on current carrying
        electrodes are discarded (like 'no_meas_current' option in EIDORS3D).

    Returns
    -------
    np.ndarray
        measurements pattern / subtract_row pairs [N, M]; shape (n_exc, n_meas_per_exc, 2)
    """
    if not isinstance(parser, list):  # transform parser into list
        parser = [parser]
    meas_current = "meas_current" in parser
    fmmu_rotate = any(p in ("fmmu", "rotate_meas") for p in parser)

    diff_op = []
    for ex_line in ex_mat:
        a, b = ex_line[0], ex_line[1]
        i0 = a if fmmu_rotate else 0
        m = (i0 + np.arange(n_el)) % n_el
        n = (m + step) % n_el
        meas_pattern = np.vstack([n, m]).T

        if not meas_current:
            diff_keep = np.logical_and.reduce((m != a, m != b, n != a, n != b))
            meas_pattern = meas_pattern[diff_keep]

        diff_op.append(meas_pattern)

    return np.array(diff_op)


def build_exc_pattern_std(n_el: int = 16, dist: int = 1) -> np.ndarray:
    """
    Generate scan matrix, `ex_mat` ( or excitation pattern), see notes

    Parameters
    ----------
    n_el : int, optional
        number of electrodes, by default 16
    dist : int, optional
        distance (number of electrodes) of A to B, by default 1
        For 'adjacent'- or 'neighbore'-mode (default) use `1` , and
        for 'apposition'-mode use `n_el/2` (see Examples).

    Returns
    -------
    np.ndarray
        stimulation matrix; shape (n_exc, 2)

    Notes
    -----
        - in the scan of EIT (or stimulation matrix), we use 4-electrodes
        mode, where A, B are used as positive and negative stimulation
        electrodes and M, N are used as voltage measurements.
        - `1` (A) for positive current injection, `-1` (B) for negative current
        sink

    Examples
    --------
        n_el=16
        if mode=='neighbore':
            ex_mat = build_exc_pattern(n_el=n_el)
        elif mode=='apposition':
            ex_mat = build_exc_pattern(dist=n_el/2)

    WARNING
    -------
        `ex_mat` is a local index, where it is ranged from 0...15, within the
        range of the number of electrodes. In FEM applications, you should
        convert `ex_mat` to global index using the (global) `el_pos` parameters.
    """
    return np.array([[i, np.mod(i + dist, n_el)] for i in range(n_el)])



# coding: utf-8
# pylint: disable=invalid-name, too-many-arguments
# pylint: disable=too-many-instance-attributes
"""
This is a python code template that guide you through
writing your own reconstruction algorithms.
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import absolute_import, division, print_function
from abc import ABC, abstractmethod
import numpy as np
from pyeit.eit.protocol import PyEITProtocol
from pyeit.mesh import PyEITMesh

from .fem import EITForward


class SolverNotReadyError(BaseException):
    """Is raised if solver.setup() not called before using solver"""


class EitBase(ABC):
    """
    Base EIT solver.
    """

    def __init__(
        self,
        mesh: PyEITMesh,
        protocol: PyEITProtocol,
    ) -> None:
        """
        An EIT solver.

        WARNING: Before using it run solver.setup() to set the solver ready!

        Parameters
        ----------
        mesh: PyEITMesh
            mesh object
        protocol: PyEITProtocol
            measurement object
        """
        # build forward solver
        self.fwd = EITForward(mesh=mesh, protocol=protocol)

        # initialize other parameters
        self.params = None
        self.xg = None
        self.yg = None
        self.mask = None
        # user must run solver.setup() manually to get correct H
        self.H = None
        self.is_ready = False

    @property
    def mesh(self) -> PyEITMesh:
        return self.fwd.mesh

    # # if needed protocol attributes can be accessed by using self.protocol
    # # instead of self.fwd.protocol
    # @property
    # def protocol(self)->PyEITProtocol:
    #     return self.fwd.protocol

    @abstractmethod
    def setup(self) -> None:
        """
        Setup EIT solver

        1. memory parameters in self.params
        2. compute some other stuff needed for 3.
        3. compute self.H used for solving inv problem by using
            >> self.H=self._compute_h()
        4. set flag self.is_ready to `True`
        """

    @abstractmethod
    def _compute_h(self) -> np.ndarray:
        """
        Compute H matrix for solving inv problem

        To be used in self.setup()
        >> self.H=self._compute_h()

        Returns
        -------
        np.ndarray
            H matrix
        """

    def solve(
        self,
        v1: np.ndarray,
        v0: np.ndarray,
        normalize: bool = False,
        log_scale: bool = False,
    ) -> np.ndarray:
        """
        Dynamic imaging (conductivities imaging)

        Parameters
        ----------
        v1: np.ndarray
            current frame
        v0: np.ndarray
            referenced frame, d = H(v1 - v0)
        normalize: Bool, optional
            true for conducting normalization, by default False
        log_scale: Bool, optional
            remap reconstructions in log scale, by default False

        Raises
        ------
            SolverNotReadyError: raised if solver not ready
            (see self._check_solver_is_ready())

        Returns
        -------
        ds: np.ndarray
            complex-valued np.ndarray, changes of conductivities
        """
        self._check_solver_is_ready()
        dv = self._normalize(v1, v0) if normalize else v1 - v0
        ds = -np.dot(self.H, dv.transpose())  # s = -Hv
        if log_scale:
            ds = np.exp(ds) - 1.0
        return ds

    def map(self, dv: np.ndarray) -> np.ndarray:
        """
        (NOT USED, Deprecated?) simple mat using projection matrix

        return -H*dv, dv should be normalized.

        Parameters
        ----------
        dv : np.ndarray
            voltage measurement frame difference (reference frame - current frame)

        Raises
        ------
        SolverNotReadyError
            raised if solver not ready (see self._check_solver_is_ready())

        Returns
        -------
        np.ndarray
            -H*dv
        """
        self._check_solver_is_ready()
        return -np.dot(self.H, dv.transpose())

    def _check_solver_is_ready(self) -> None:
        """
        Check if solver is ready for solving

        Addtionaly test also if self.H not `None`

        Raises
        ------
        SolverNotReadyError
            raised if solver not ready
        """
        if not self.is_ready or self.H is None:
            msg = "User must first run {type(self).__name__}.setup() before imaging."
            raise SolverNotReadyError(msg)

    def _normalize(self, v1: np.ndarray, v0: np.ndarray) -> np.ndarray:
        """
        Normalize current frame using the amplitude of the reference frame.
        Boundary measurements v are complex-valued, we can use the real part of v,
        np.real(v), or the absolute values of v, np.abs(v).

        Parameters
        ----------
        v1 : np.ndarray
            current frame, can be a Nx192 matrix where N is the number of frames
        v0 : np.ndarray
            referenced frame, which is a row vector

        Returns
        -------
        np.ndarray
            Normalized current frame difference dv
        """
        return (v1 - v0) / np.abs(v0)



""" main module for 2D/3D mesh """
from .wrapper import PyEITMesh, create, set_perm, layer_circle
from .shell import multi_shell, multi_circle
from .utils import check_ccw

__all__ = [
    "PyEITMesh",
    "create",
    "set_perm",
    "layer_circle",
    "multi_shell",
    "multi_circle",
    "check_ccw",
]



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-instance-attributes
# pylint: disable=too-many-arguments, arguments-differ
"""
GREIT (using distribution method)

Note, that, the advantages of greit is NOT on simulated data, but
1. construct RM using real-life data with a stick move in the cylinder
2. construct RM on finer mesh, and use coarse-to-fine map for visualization
3. more robust to noise by adding noise via (JJ^T + lamb*Sigma_N)^{-1}
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import absolute_import, division, print_function

from typing import Tuple

import numpy as np
import scipy.linalg as la

from .base import EitBase
from .interp2d import meshgrid, weight_sigmod


class GREIT(EitBase):
    """The GREIT algorithm"""

    def setup(
        self,
        method: str = "dist",
        w: np.ndarray = None,
        p: float = 0.20,
        lamb: float = 1e-2,
        n: int = 32,
        s: float = 20.0,
        ratio: float = 0.1,
        jac_normalized: bool = False,
    ) -> None:
        """
        Setup GREIT solver

        Parameters
        ----------
        method : str, optional
            only 'dist' accepted, by default "dist"
        w : np.ndarray, optional
            weight on each element, by default None
        p : float, optional
            noise covariance, by default 0.20
        lamb : float, optional
            regularization parameters, by default 1e-2
        n : int, optional
            grid size, by default 32
        s : float, optional
            control the blur, by default 20.0
        ratio : float, optional
            desired ratio, by default 0.1
        jac_normalized : bool, optional
            normalize the jacobian using f0 computed from input perm, by
            default False

        Raises
        ------
        ValueError
            raised if method != "dist"

        References
        ----------
        [1] Bartlomiej Grychtol, Beat Muller, Andy Adler
            "3D EIT image reconstruction with GREIT"
        [2] Adler, Andy, et al.
            "GREIT: a unified approach to 2D linear EIT reconstruction of
            lung images." Physiological measurement 30.6 (2009): S35.
        """

        if method != "dist":
            raise ValueError(f"method {method} not supported yet")

        # parameters for GREIT projection
        if w is None:
            w = np.ones_like(self.mesh.perm)
        self.params = {
            "w": w,
            "p": p,
            "lamb": lamb,
            "n": n,
            "s": s,
            "ratio": ratio,
            "jac_normalize": jac_normalized,
        }

        # Build grids and mask
        self.xg, self.yg, self.mask = meshgrid(self.mesh.node, n=n)

        w_mat = self._compute_grid_weights(self.xg, self.yg)
        self.J, self.v0 = self.fwd.compute_jac(normalize=jac_normalized)
        self.H = self._compute_h(jac=self.J, w_mat=w_mat)
        self.is_ready = True

    def _compute_h(self, jac: np.ndarray, w_mat: np.ndarray) -> np.ndarray:
        """
        Generate H (or R) using distribution method for GREIT solver

        Args:
            jac (np.ndarray): Jacobian matrix
            w_mat (np.ndarray): meights matrix

        Returns:
            np.ndarray: H
        """
        lamb, p = self.params["lamb"], self.params["p"]
        # E[yy^T], it is more efficient to use left pinv than right pinv
        j_j_w = np.dot(jac, jac.T)
        r_mat = np.diag(np.diag(j_j_w) ** p)
        jac_inv = la.inv(j_j_w + lamb * r_mat)
        # RM = E[xx^T] / E[yy^T]
        return np.dot(np.dot(w_mat.T, jac.T), jac_inv)

    def get_grid(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Return masking grid data

        Raises
        ------
        SolverNotReadyError
            raised if solver not ready (see self._check_solver_is_ready())

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            x grid, y grid and masking data, which denotes nodes outside
            2D mesh
        """
        self._check_solver_is_ready()
        return self.xg, self.yg, self.mask

    def mask_value(
        self, ds: np.ndarray, mask_value: float = 0
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Set mask values on nodes outside 2D mesh. (for plot only)

        Parameters
        ----------
        ds : np.ndarray
            conductivity data on nodes
        mask_value : float, optional
            mask conductivity value to set on nodes outside 2D mesh, by
            default 0

        Raises
        ------
        SolverNotReadyError
            raised if solver not ready (see self._check_solver_is_ready())

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            x grid, y grid and "masked" conductivity data on nodes
        """
        self._check_solver_is_ready()
        ds[self.mask] = mask_value
        ds = ds.reshape(self.xg.shape)
        return self.xg, self.yg, ds

    def _compute_grid_weights(self, xg: np.ndarray, yg: np.ndarray) -> np.ndarray:
        """
        Compute weights for given grid (xg,yg)

        Parameters
        ----------
        xg : np.ndarray
            x grid
        yg : np.ndarray
            y grid

        Returns
        -------
        np.ndarray
            weights
        """
        # mapping from values on triangles to values on grids
        xy = (
            self.mesh.elem_centers
        )  # np.mean(self.mesh.node[self.mesh.element], axis=1)
        xyi = np.vstack((xg.flatten(), yg.flatten())).T
        # GREIT is using sigmod as weighting function (global)
        ratio, s = self.params["ratio"], self.params["s"]
        return weight_sigmod(xy, xyi, ratio=ratio, s=s)

    @staticmethod
    def build_set(x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """generate R from a set of training sets (deprecate)."""
        # E_w[yy^T]
        y_y_t = la.inv(np.dot(y, y.transpose()))
        return np.dot(np.dot(x, y), y_y_t)



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-locals, no-name-in-module
""" interpolation on 2D/3D irregular/regular grids """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import absolute_import, division, print_function

from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg as la
from matplotlib.path import Path

# for debugging
from pyeit.mesh import layer_circle, set_perm
from scipy.sparse import coo_matrix
from scipy.spatial import ConvexHull


def meshgrid(
    pts: np.ndarray, n: int = 32, ext_ratio: float = 0.0, gc: bool = False
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    build xg, yg, mask grids from triangles point cloud
    function for interpolating regular grids

    Parameters
    ----------
    pts: np.ndarray
        nx2 array of points (x, y)
    n: int
        the number of meshgrid per dimension, by default 32
    ext_ratio: float
        extend the boundary of meshgrid by ext_ratio*d, by default 0.0
    gc: bool
        grid_correction, offset xgrid and ygrid by half step size , by default
        False

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]
        x grid, y grid, mask

    Notes
    -----
    mask denotes points outside mesh.
    """
    xg, yg = _build_grid(pts, n=n, ext_ratio=ext_ratio, gc=gc)
    pts_edges = _hull_points(pts)
    mask = _build_mask(pts_edges, xg, yg)
    return xg, yg, mask


def _build_grid(
    pts: np.ndarray, n: int = 32, ext_ratio: float = 0.0, gc: bool = False
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generating mesh grid from triangles point cloud

    Parameters
    ----------
    pts: np.ndarray
        nx2 array of points (x, y)
    n: int
        the number of meshgrid per dimension, by default 32
    ext_ratio: float
        extend the boundary of meshgrid by ext_ratio*d, by default 0.0
    gc: bool
        grid_correction, offset xgrid and ygrid by half step size , by default
        False

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        x grid, y grid
    """
    x, y = pts[:, 0], pts[:, 1]
    x_min, x_max = min(x), max(x)
    y_min, y_max = min(y), max(y)
    x_ext = (x_max - x_min) * ext_ratio
    y_ext = (y_max - y_min) * ext_ratio
    xv, xv_step = np.linspace(
        x_min - x_ext, x_max + x_ext, num=n, endpoint=False, retstep=True
    )
    yv, yv_step = np.linspace(
        y_min - y_ext, y_max + y_ext, num=n, endpoint=False, retstep=True
    )
    # if need grid correction
    if gc:
        xv = xv + xv_step / 2.0
        yv = yv + yv_step / 2.0
    xg, yg = np.meshgrid(xv, yv, sparse=False, indexing="xy")
    return xg, yg


def _build_mask(pts_edges: np.ndarray, xg: np.ndarray, yg: np.ndarray) -> np.ndarray:
    """
    find whether meshgrids is interior of mesh

    Parameters
    ----------
    pts_edges : np.ndarray
        points on the edges of the mesh
    xg : np.ndarray
        x grid
    yg : np.ndarray
        x grid

    Returns
    -------
    np.ndarray
        mask (denotes points outside mesh.)
    """
    # 1. create mask based on meshes
    points = np.vstack((xg.flatten(), yg.flatten())).T

    # 2. extract edge points using el_pos
    path = Path(pts_edges, closed=False)
    mask = path.contains_points(points)
    return ~mask


def _hull_points(pts: np.ndarray) -> np.ndarray:
    """
    return the convex hull points from a point cloud

    Parameters
    ----------
    pts: np.ndarray
        nx2 array of points (x, y) (can be also (x,y,z))

    Returns
    -------
    np.ndarray
        convex hull points (edge points)
    """
    pts_2D = pts[:, :2]  # get only x and y
    cv = ConvexHull(pts_2D)
    hull_nodes = cv.vertices
    return pts_2D[hull_nodes, :]


def _distance2d(
    x: np.ndarray, y: np.ndarray, center: Tuple[str, list] = "mean"
) -> np.ndarray:
    """
    Calculate radius given center.
    This function can be OPTIMIZED using numba or cython.

    Parameters
    ----------
    x : np.ndarray
        nx1 array of x coordiante
    y : np.ndarray
        nx1 array of y coordiante
    center : Tuple[str, list], optional
        center definition, by default "mean".
        If center is `None`, [0,0] will be used.
        If center is "mean", [np.mean(x), np.mean(y)] will be used.
        If center is list, [center[0], center[1]] will be used.

    Returns
    -------
    np.ndarray
        nx1 array of distance from center to points (x,y)
    """
    if center is None:
        xc, yc = 0, 0
    elif center == "mean":
        xc, yc = np.mean(x), np.mean(y)
    else:
        xc, yc = center[0], center[1]
    # distance 2d
    return np.sqrt((x - xc) ** 2 + (y - yc) ** 2).ravel()


def _distance_matrix2d(xy: np.ndarray, xyi: np.ndarray) -> np.ndarray:
    """
    (2D only) return element-wise distance matrix (pair-wise)

    Parameters
    ----------
    xy : np.ndarray
        nx2 array of points (x, y)
    xyi : np.ndarray
        points pairs

    Returns
    -------
    np.ndarray
        distance matrix between pairwise observations
    """
    # Make a distance matrix between pairwise observations
    # Note: from <http://stackoverflow.com/questions/1871536>
    # (Yay for ufuncs!)
    d0 = np.subtract.outer(xy[:, 0], xyi[:, 0])  # size(xy) * size(xyi)
    d1 = np.subtract.outer(xy[:, 1], xyi[:, 1])

    # hypot : element-wise sqrt(d0**2 + d1**2)
    return np.hypot(d0, d1)


def weight_sigmod(
    xy: np.ndarray, xyi: np.ndarray, ratio: float = 0.05, s: float = 20.0
) -> np.ndarray:
    """
    (2D only)
    local weight/interpolate by sigmod function (GREIT3D)

    Parameters
    ----------
    xy: np.ndarray
        (x, y) of values
    xyi: np.ndarray
        (xi, yi) of interpolated locations
    ratio: float
        R0 = d_max * ratio, by default 0.05.
    s: float
        control the decay ratio, by default 20.0.

    Returns
    -------
    w_mat: np.ndarray
        weighting matrix mapping from xy to xyi (xy meshgrid)
    """
    d_mat = _distance_matrix2d(xy, xyi)
    # normalize distance
    d_max = np.max(d_mat)
    d_mat = 5.0 * d_mat / d_max
    # desired radius (a ratio of max pairwise distance)
    r0 = 5.0 * ratio
    # weights is the sigmod function
    weight = 1.0 / (1 + np.exp(s * (d_mat - r0)))
    # weighting matrix normalized
    return weight / weight.sum(axis=0)


def weight_idw(
    xy: np.ndarray, xyi: np.ndarray, k: int = 6, p: float = 1.0
) -> np.ndarray:
    """
    (2D only)
    local weight/interpolate by inverse distance

    Parameters
    ----------
    xy: np.ndarray
        (x, y) of values
    xyi: np.ndarray
        (xi, yi) of interpolated locations
    k: int
        number of nearest neighbores, by default 6.
    p: float
        scaling distance, by default 1.0.

    Returns
    -------
    w_mat: np.ndarray
        weighting matrix mapping from xy to xy_mesh
    """
    d_mat = _distance_matrix2d(xy, xyi)
    # weight = 1.0 / d_mat**p
    weight = 1.0 / d_mat**p
    # keep only k largest neighbores (nearest)
    for w in weight.T:
        sort_indices = np.argsort(w)
        np.put(w, sort_indices[:-k], 0)
    # weighting matrix normalized
    # xy times xyi size, use w_mat.T to multiply
    return weight / weight.sum(axis=0)


def weight_linear_rbf(xy: np.ndarray, xyi: np.ndarray, z: np.ndarray) -> np.ndarray:
    """
    (2D only)
    local weight/interpolate by linear rbf function (z value required)

    Parameters
    ----------
    xy: np.ndarray
        (x, y) of values
    xyi: np.ndarray
        (xi, yi) of interpolated locations
    z: np.ndarray
        z values

    Returns
    -------
    w_mat: np.ndarray
        weighting matrix mapping from xy to xy_mesh
    """
    internal_dist = _distance_matrix2d(xy, xy)
    weights = la.solve(internal_dist, z)
    interp_dist = _distance_matrix2d(xy, xyi)
    return np.dot(interp_dist.T, weights)


def weight_barycentric_gradient():
    """
    (2D only)
    local weight/interpolate by barycentric gradient

    Parameters
    ----------
    xy: np.ndarray
        (x, y) of values
    xyi: np.ndarray
        (xi, yi) of interpolated locations

    Returns
    -------
    w_mat: np.ndarray
        weighting matrix mapping from xy to xy_mesh
    """
    raise NotImplementedError()


def sim2pts(pts: np.ndarray, sim: np.ndarray, sim_values: np.ndarray) -> np.ndarray:
    """
    (2D/3D) compatible.

    Interp values on points using values on simplex,
    a simplex can be triangle or tetrahedron.
    The areas/volumes are used as weights.

    f_n = (sum_e r_e*S_e) / (sum_e S_e)

    where r_e is the value on triangles who share the node n,
    S_e is the area of triangle e.

    Parameters
    ----------
    pts_values: np.ndarray
        Nx1 array, real/complex valued
    sim: np.ndarray
        Mx3, Mx4 array, elements or simplex
        triangles denote connectivity [[i, j, k]]
        tetrahedrons denote connectivity [[i, j, m, n]]
    sim_value: np.ndarray

    Notes
    -----
    This function is similar to pdeprtni of MATLAB pde.
    """
    N = pts.shape[0]
    M, dim = sim.shape
    # calculate the weights
    # triangle/tetrahedron must be CCW (recommended), then a is positive
    if dim == 3:
        weight_func = tri_area
    elif dim == 4:
        weight_func = tet_volume
    weights = weight_func(pts, sim)
    # build tri->pts matrix, could be accelerated using sparse matrix
    row = np.ravel(sim)
    col = np.repeat(np.arange(M), dim)  # [0, 0, 0, 1, 1, 1, ...]
    data = np.repeat(weights, dim)
    e2n_map = coo_matrix((data, (row, col)), shape=(N, M)).tocsr()
    # map values from elements to nodes
    # and re-weight by the sum of the areas/volumes of adjacent elements
    f = e2n_map.dot(sim_values)
    w = np.sum(e2n_map.toarray(), axis=1)

    return f / w


def pts2sim(sim: np.ndarray, pts_values: np.ndarray) -> np.ndarray:
    """
    (2D/3D) compatible.

    Given values on nodes, calculate interpolated values on simplex,
    this function was tested and equivalent to MATLAB 'pdeintrp'
    except for the shapes of 'pts' and 'tri'

    Parameters
    ----------
    sim: np.ndarray
        Mx3, Mx4 array, elements or simplex
        triangles denote connectivity [[i, j, k]]
        tetrahedrons denote connectivity [[i, j, m, n]]
    pts_values: np.ndarray
        Nx1 array, real/complex valued

    Returns
    -------
    el_value: np.ndarray
        Mx1 array, real/complex valued

    Notes
    -----
    This function is similar to pdfinterp of MATLAB pde.
    """
    # averaged over 3 nodes of a triangle
    return np.mean(pts_values[sim], axis=1)


def tri_area(pts: np.ndarray, sim: np.ndarray) -> np.ndarray:
    """
    calculate the area of each triangle

    Parameters
    ----------
    pts: np.ndarray
        Nx2 array, (x,y) locations for points (can be also (x,y,z))
    sim: np.ndarray
        Mx3 array, elements (triangles) connectivity

    Returns
    -------
    a: np.ndarray
        Areas of triangles
    """
    pts_2D = pts[:, :2]  # get only x and y
    a = np.zeros(np.shape(sim)[0])
    for i, e in enumerate(sim):
        xy = pts_2D[e]
        # which can be simplified to
        # s = xy[[2, 0, 1]] - xy[[1, 2, 0]]
        s = xy[[2, 0]] - xy[[1, 2]]

        # a should be positive if triangles are CCW arranged
        a[i] = la.det(s)

    return a * 0.5


def tet_volume(pts: np.ndarray, sim: np.ndarray) -> np.ndarray:
    """
    calculate the area of each triangle

    Parameters
    ----------
    pts: np.ndarray
        Nx3 array, (x,y, z) locations for points
    sim: np.ndarray
        Mx4 array, elements (tetrahedrons) connectivity

    Returns
    -------
    v: np.ndarray
        Volumes of tetrahedrons
    """
    v = np.zeros(np.shape(sim)[0])
    for i, e in enumerate(sim):
        xyz = pts[e]
        s = xyz[[2, 3, 0]] - xyz[[1, 2, 3]]

        # a should be positive if triangles are CCW arranged
        v[i] = la.det(s)

    return v / 6.0


def pdetrg(pts: np.ndarray, tri: np.ndarray) -> np.ndarray:
    """
    (Deprecated)
    analytical calculate the Area and grad(phi_i) using
    barycentric coordinates (simplex coordinates)
    this function is tested and equivalent to MATLAB 'pdetrg'
    except for the shape of 'pts' and 'tri' and the output

    note: each node may have multiple gradients in neighbor
    elements' coordinates. you may averaged all the gradient to
    get one node gradient.

    Parameters
    ----------
    pts: np.ndarray
        Nx2 array, (x,y) locations for points
    tri: np.ndarray
        Mx3 array, elements (triangles) connectivity

    Returns
    -------
    a: np.ndarray
        Mx1 array, areas of elements
    grad_phi_x: np.ndarray
        Mx3 array, x-gradient on elements' local coordinate
    grad_phi_y: np.ndarray
        Mx3 array, y-gradient on elements' local coordinate
    """
    m = np.size(tri, 0)

    ix = tri[:, 0]
    iy = tri[:, 1]
    iz = tri[:, 2]

    s1 = pts[iz, :] - pts[iy, :]
    s2 = pts[ix, :] - pts[iz, :]
    s3 = pts[iy, :] - pts[ix, :]

    a = 0.5 * (s2[:, 0] * s3[:, 1] - s3[:, 0] * s2[:, 1])
    if any(a) < 0:
        raise ValueError("Triangles are not in CCW order")

    # note in python, reshape place elements first on the right-most index
    grad_phi_x = np.reshape(
        [-s1[:, 1] / (2.0 * a), -s2[:, 1] / (2.0 * a), -s3[:, 1] / (2.0 * a)], [-1, m]
    ).T
    grad_phi_y = np.reshape(
        [s1[:, 0] / (2.0 * a), s2[:, 0] / (2.0 * a), s3[:, 0] / (2.0 * a)], [-1, m]
    ).T

    return a, grad_phi_x, grad_phi_y


def pdegrad(
    pts: np.ndarray, tri: np.ndarray, node_value: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    (Deprecated)
    given values on nodes, calculate the averaged-grad on elements
    this function was tested and equivalent to MATLAB 'pdegrad'
    except for the shape of 'pts', 'tri'

    Parameters
    ----------
    pts: np.ndarray
        Nx2 array, (x,y) locations for points
    tri: np.ndarray
        Mx3 array, elements (triangles) connectivity
    node_value: np.ndarray
        Nx1 array, real/complex valued

    Returns
    -------
    el_grad: np.ndarray
        el_grad, Mx2 array, real/complex valued
    """
    m = np.size(tri, 0)
    _, grad_phi_x, grad_phi_y = pdetrg(pts, tri)
    tri_values = np.reshape(node_value[tri.ravel()], [m, -1])
    grad_el_x = np.sum(grad_phi_x * tri_values, axis=1)
    grad_el_y = np.sum(grad_phi_y * tri_values, axis=1)
    return grad_el_x, grad_el_y


def demo() -> None:
    from pyeit.mesh.wrapper import PyEITAnomaly_Circle

    """demo shows how to interpolate on regular/irregular grids"""
    # 1. create mesh
    mesh_obj = layer_circle(n_layer=8, n_fan=6)
    pts = mesh_obj.node
    tri = mesh_obj.element

    # set anomaly
    anomaly = PyEITAnomaly_Circle(center=[0.5, 0.5], r=0.2, perm=100.0)
    mesh_new = set_perm(mesh_obj, anomaly=anomaly)

    # 2. interpolate using averaged neighbor triangle area
    perm_node = sim2pts(pts, tri, mesh_new.perm)

    # plot mesh and interpolated mesh (tri2pts)
    fig_size = (6, 4)
    fig = plt.figure(figsize=fig_size)
    ax = fig.add_subplot(111)
    ax.set_aspect("equal")
    ax.triplot(pts[:, 0], pts[:, 1], tri)
    im1 = ax.tripcolor(pts[:, 0], pts[:, 1], tri, mesh_new.perm)
    fig.colorbar(im1, orientation="vertical")

    fig = plt.figure(figsize=fig_size)
    ax2 = fig.add_subplot(111)
    ax2.set_aspect("equal")
    ax2.triplot(pts[:, 0], pts[:, 1], tri)
    im2 = ax2.tripcolor(pts[:, 0], pts[:, 1], tri, perm_node, shading="flat")
    fig.colorbar(im2, orientation="vertical")

    # 3. interpolate on grids (irregular or regular) using IDW, sigmod
    xg, yg, mask = meshgrid(pts)
    im = np.ones_like(mask)
    # mapping from values on xy to values on xyi
    xy = np.mean(pts[tri], axis=1)
    xyi = np.vstack((xg.flatten(), yg.flatten())).T
    # w_mat = weight_idw(xy, xyi)
    w_mat = weight_sigmod(xy, xyi)
    im = np.dot(w_mat.T, mesh_new.perm)
    # im = weight_linear_rbf(xy, xyi, mesh_new['perm'])
    im[mask] = 0.0
    # reshape to grid size
    im = im.reshape(xg.shape)

    # plot interpolated values
    fig, ax = plt.subplots(figsize=fig_size)
    ax.set_aspect("equal")
    ax.triplot(pts[:, 0], pts[:, 1], tri, alpha=0.5)
    im3 = ax.pcolor(xg, yg, im, edgecolors=None, linewidth=0, alpha=0.8)
    fig.colorbar(im3, orientation="vertical")
    plt.show()


if __name__ == "__main__":
    demo()



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-arguments
""" wrapper function of distmesh for EIT """
# Copyright (c) Benyuan Liu. All rights reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import absolute_import, division, print_function

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Union

import numpy as np

from . import shape
from .distmesh import build
from .mesh_circle import MeshCircle
from .shape import ball, circle
from .utils import check_order


@dataclass
class PyEITMesh:
    """
    Pyeit buid-in mesh object

    perm : Union[int, float, np.ndarray], optional
        permittivity on elements ; shape (n_elems,), by default `None`.
        If `None`, a uniform permittivity on elements with a value 1 will be generated.
        If perm is int or float, uniform permittivity on elements with value of perm  will be generated.
    """

    element: np.ndarray  # elements of the mesh of shape (n_elem, 3) for 2D mesh, (n_elem, 4) for 3D mesh
    node: np.ndarray  # node of the mesh of shape (n_nodes, 2), (n_nodes, 3)
    el_pos: np.ndarray  # node corresponding to each electrodes of shape (n_el, 1)
    perm: Union[int, float, np.ndarray] = None  # permittivity on element
    ref_el: int = None  # node of the reference electrode # ref node should not be on electrodes, it is up to the user to decide

    def __post_init__(self) -> None:
        """Checking of the inputs"""
        self.element = self._check_element(self.element)
        self.node = self._check_node(self.node)
        self.perm = self.get_valid_perm(self.perm)
        self.ref_el = self._check_ref_el(self.ref_el)

    def print_stats(self):
        """
        Print mesh or tetrahedral status

        Parameters
        ----------
        p: array_like
            coordinates of nodes (x, y) in 2D, (x, y, z) in 3D
        t: array_like
            connectives forming elements

        Notes
        -----
        a simple function for illustration purpose only.
        print the status (size) of nodes and elements
        """
        text_2D_3D = "3D" if self.is_3D else "2D"
        print(f"{text_2D_3D} mesh status:")
        print(f"{self.n_nodes} nodes, {self.n_elems} elements")

    def _check_element(self, element: np.ndarray) -> np.ndarray:
        """
        Check nodes element
        return nodes [x,y,z]

        Parameters
        ----------
        node : np.ndarray, optional
            nodes [x,y] ; shape (n_elem,3)
            nodes [x,y,z] ; shape (n_nodes,4)

        Returns
        -------
        np.ndarray
            nodes [x,y,z] ; shape (n_nodes,3)

        Raises
        ------
        TypeError
            raised if perm is not ndarray and of shape (n_tri,)
        """
        if not isinstance(element, np.ndarray):
            raise TypeError(f"Wrong type of {element=}, expected an ndarray")
        if element.ndim != 2:
            raise TypeError(
                f"Wrong shape of {element.shape=}, expected an ndarray with 2 dimensions"
            )
        if element.shape[1] not in [3, 4]:
            raise TypeError(
                f"Wrong shape of {element.shape=}, expected an ndarray of shape (n_nodes,2) or (n_nodes,3)"
            )

        return element

    def _check_node(self, node: np.ndarray) -> np.ndarray:
        """
        Check nodes shape
        return nodes [x,y,z]

        Parameters
        ----------
        node : np.ndarray, optional
            nodes [x,y] ; shape (n_nodes,2) (in that case z will be set 0)
            nodes [x,y,z] ; shape (n_nodes,3)

        Returns
        -------
        np.ndarray
            nodes [x,y,z] ; shape (n_nodes,3)

        Raises
        ------
        TypeError
            raised if perm is not ndarray and of shape (n_tri,)
        """
        if not isinstance(node, np.ndarray):
            raise TypeError(f"Wrong type of {node=}, expected an ndarray")
        if node.ndim != 2:
            raise TypeError(
                f"Wrong shape of {node.shape=}, expected an ndarray with 2 dimensions"
            )
        if node.shape[1] not in [2, 3]:
            raise TypeError(
                f"Wrong shape of {node.shape=}, expected an ndarray of shape (n_nodes,2) or (n_nodes,3)"
            )
        # convert nodes [x,y] to nodes [x,y,0]
        if node.shape[1] == 2:
            node = np.hstack((node, np.zeros((node.shape[0], 1))))

        return node

    def get_valid_perm(self, perm: Union[int, float, np.ndarray] = None) -> np.ndarray:
        """
        Return a valid permittivity on element

        Parameters
        ----------
        perm : Union[int, float, np.ndarray], optional
            Permittivity on elements ; shape (n_elems,), by default `None`.
            If `None`, a uniform permittivity on elements with a value 1 will be used.
            If perm is int or float, uniform permittivity on elements will be used.

        Returns
        -------
        np.ndarray
            permittivity on elements ; shape (n_elems,)

        Raises
        ------
        TypeError
            check if perm passed as ndarray has a shape (n_elems,) # TODO enhanced
        """

        if perm is None:
            return np.ones(self.n_elems, dtype=float)
        elif isinstance(perm, (int, float)):
            return np.ones(self.n_elems, dtype=float) * perm

        if not isinstance(perm, np.ndarray) or perm.shape != (self.n_elems,):
            raise TypeError(
                f"Wrong type/shape of {perm=}, expected an ndarray; shape ({self.n_elems}, )"
            )
        return perm

    def _check_ref_el(self, ref_el: int = None) -> int:
        """
        Return a valid reference electrode node

        Parameters
        ----------
        ref_el : int, optional
            node number of reference electrode, by default None
            If None, a default node will be assigned
            If the choosen node is on electrode node, a default node will be assigned

        eturns
        -------
        int
            valid reference electrode node

        """
        return (
            ref_el
            if ref_el is not None and ref_el not in self.el_pos
            else max(self.el_pos) + 1
        )

    def set_ref_el(self, ref_el: int) -> None:
        """
        Set reference electrode node

        Parameters
        ----------
        ref_el : int, optional
            node number of reference electrode
            If the choosen node is on electrode node, a default node will be assigned

        """
        self.ref_el = self._check_ref_el(ref_el)

    @property
    def n_nodes(self) -> int:
        """
        Returns
        -------
        int
            number of nodes contained in the mesh
        """
        return self.node.shape[0]

    @property
    def n_elems(self) -> int:
        """
        Returns
        -------
        int
            number of elements contained in the mesh
        """
        return self.element.shape[0]

    @property
    def n_vertices(self) -> int:
        """
        Returns
        -------
        int
            number of vertices of the elements contained in the mesh
        """
        return self.element.shape[1]

    @property
    def n_el(self) -> int:
        """
        Returns
        -------
        int
            number of electrodes
        """
        return self.el_pos.shape[0]

    @property
    def elem_centers(self) -> np.ndarray:
        """
        Returns
        -------
        np.ndarray
            center of the nodes [x,y,z]; shape (n_elems,3)
        """
        return np.mean(self.node[self.element], axis=1)

    @property
    def is_3D(self) -> np.ndarray:
        """
        Returns
        -------
        np.ndarray
            True if the mesh is a 3D mesh (use elements with 4 vertices)
        """
        return self.n_vertices == 4

    @property
    def is_2D(self) -> np.ndarray:
        """
        Returns
        -------
        np.ndarray
            True if the mesh is a 2D mesh (use elements with 3 vertices)
        """
        return self.n_vertices == 3


def create(
    n_el: int = 16,
    fd: Callable = None,
    fh: Callable = shape.area_uniform,
    h0: float = 0.1,
    p_fix: np.ndarray = None,
    bbox: np.ndarray = None,
) -> PyEITMesh:
    """
    Generating 2D/3D meshes using distmesh (pyEIT built-in)

    Parameters
    ----------
    n_el: int
        number of electrodes (point-type electrode)
    fd: function
        distance function (circle in 2D, ball in 3D)
    fh: function
        mesh size quality control function
    p_fix: NDArray
        fixed points
    bbox: NDArray
        bounding box
    h0: float
        initial mesh size, default=0.1

    Returns
    -------
    PyEITMesh
        mesh object
    """

    # test conditions if fd or/and bbox are none

    if bbox is None:
        if fd != shape.ball:
            bbox = np.array([[-1, -1], [1, 1]])
        else:
            bbox = [[-1.2, -1.2, -1.2], [1.2, 1.2, 1.2]]

    # list is converted to Numpy array so we can use it then (calling shape method..)
    bbox = np.array(bbox)
    n_dim = bbox.shape[1]  # bring dimension

    # infer dim
    if fd is None:
        if n_dim == 2:
            fd = shape.circle
        elif n_dim == 3:
            fd = shape.ball

    if n_dim not in [2, 3]:
        raise TypeError("distmesh only supports 2D or 3D")
    if bbox.shape[0] != 2:
        raise TypeError("please specify lower and upper bound of bbox")

    if p_fix is None:
        if n_dim == 2:
            if fd == shape.thorax:
                p_fix = shape.thorax_pfix
            elif fd == shape.head_symm:
                p_fix = shape.head_symm_pfix
            elif fd == shape.lshape:
                p_fix = shape.lshape_pfix
                h0 = 0.15
            else:
                p_fix = shape.fix_points_fd(fd, n_el=n_el)
        elif n_dim == 3:
            p_fix = shape.fix_points_ball(n_el=n_el)

    # 1. build mesh
    p, t = build(fd, fh, pfix=p_fix, bbox=bbox, h0=h0)
    # 2. check whether t is counter-clock-wise, otherwise reshape it
    t = check_order(p, t)
    # 3. generate electrodes, the same as p_fix (top n_el)
    el_pos = np.arange(n_el)
    return PyEITMesh(element=t, node=p, perm=None, el_pos=el_pos)


@dataclass
class PyEITAnomaly(ABC):
    """
    Pyeit Anomaly for simulation purpose
    """

    center: Union[np.ndarray, list]  # center of the anomaly
    perm: float = 1.0  # permittivity of the anomaly

    def __post_init__(self):
        if isinstance(self.center, list):
            self.center = np.array(self.center)

    @abstractmethod
    def mask(self, pts: np.ndarray) -> np.ndarray:
        """
        Return mask corresponding to the pts contained in the Anomaly
        """


@dataclass
class PyEITAnomaly_Circle(PyEITAnomaly):
    """
    Pyeit Anomaly for simulation purpose, 2D circle

    """

    r: float = 1.0  # radius of the circle

    def mask(self, pts: np.ndarray) -> np.ndarray:
        pts = pts[:, :2].reshape((-1, 2))
        pc = self.center[:2].reshape((1, 2))
        return circle(pts, pc, self.r) < 0


@dataclass
class PyEITAnomaly_Ball(PyEITAnomaly):
    """
    Pyeit Anomaly for simulation purpose, 3D ball
    """

    r: float = 1.0  # radius of the ball

    def mask(self, pts: np.ndarray) -> np.ndarray:
        pts = pts.reshape((-1, 3))
        pc = self.center.reshape((1, 3))
        return ball(pts, pc, self.r) < 0


def set_perm(
    mesh: PyEITMesh,
    anomaly: Union[PyEITAnomaly, list[PyEITAnomaly]] = None,
    background: float = None,
) -> PyEITMesh:
    """wrapper for pyEIT interface

    Note
    ----
    update permittivity of mesh, if specified.

    Parameters
    ----------
    mesh: PyEITMesh
        mesh object
    anomaly: Union[PyEITAnomaly, list[PyEITAnomaly]], optional
        anomaly object or list of anomalyobject contains,
        all permittivity on triangles whose distance to (x,y) are less than (d)
        will be replaced with a new value, 'perm' may be a complex value.
    background: float, optional
        set background permittivity

    Returns
    -------
    PyEITMesh
        mesh object
    """
    perm = mesh.perm.copy()
    # reset background if needed
    if background is not None:
        perm = background * np.ones_like(mesh.perm)

    # change dtype to 'complex' for complex-valued permittivity
    if anomaly is None:
        return mesh

    if isinstance(anomaly, PyEITAnomaly):
        anomaly = [anomaly]

    for an in anomaly:
        if np.iscomplex(an.perm):
            perm = perm.astype("complex")
            break

    # assign anomaly values (for elements in regions)
    tri_centers = mesh.elem_centers
    for an in anomaly:
        mask = an.mask(tri_centers)
        perm[mask] = an.perm

    return PyEITMesh(
        element=mesh.element,
        node=mesh.node,
        perm=perm,
        el_pos=mesh.el_pos,
        ref_el=mesh.ref_el,
    )


def layer_circle(n_el: int = 16, n_fan: int = 8, n_layer: int = 8) -> PyEITMesh:
    """generate mesh on unit-circle"""
    model = MeshCircle(n_fan=n_fan, n_layer=n_layer, n_el=n_el)
    pts, tri, el_pos = model.create()
    # perm = np.ones(tri.shape[0]) not need anymore as handled in PyEITMesh
    return PyEITMesh(element=tri, node=pts, perm=None, el_pos=el_pos)


if __name__ == "__main__":
    """"""
    PyEITAnomaly_Circle()
